/* jcifs smb client library in Java
 * Copyright (C) 2002  "Michael B. Allen" <jcifs at samba dot org>
 *                   "Jason Pugsley" <jcifs at samba dot org>
 *                   "skeetz" <jcifs at samba dot org>
 *                   "Eric Glass" <jcifs at samba dot org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.http;

import java.io.File;
import java.io.IOException;

import java.net.MalformedURLException;
import java.net.URL;

import java.security.Principal;

import java.util.Enumeration;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import jcifs.Config;
import jcifs.UniAddress;

import jcifs.netbios.NbtAddress;

import jcifs.smb.NtlmPasswordAuthentication;
import jcifs.smb.SmbAuthException;
import jcifs.smb.SmbSession;

import jcifs.util.Base64;

public class AuthenticationFilter implements Filter {

    public static final String HTTP_NEGOTIATE = "Negotiate";

    public static final String HTTP_NTLM = "NTLM";

    public static final String HTTP_BASIC = "Basic";

    private FilterConfig filterConfig;

    private String defaultDomain;

    private String domainController;

    private boolean loadBalance;

    private boolean enableNegotiate;

    private boolean enableBasic;

    private boolean insecureBasic;

    private String realm;

    public void init(FilterConfig filterConfig) throws ServletException {
        this.filterConfig = filterConfig;
        Config.setProperty( "jcifs.smb.client.soTimeout", "300000" );
        Config.setProperty( "jcifs.netbios.cachePolicy", "600" );
        Enumeration parameters = filterConfig.getInitParameterNames();
        while (parameters.hasMoreElements()) {
            String parameter = (String) parameters.nextElement();
            if (parameter.startsWith("jcifs.")) {
                Config.setProperty(parameter,
                        filterConfig.getInitParameter(parameter));
            }
        }
        if (Boolean.valueOf(filterConfig.getInitParameter(
                "sun.security.krb5.debug")).booleanValue()) {
            System.setProperty("sun.security.krb5.debug", "true");
        }
        String loginConfig = filterConfig.getInitParameter(
                "java.security.auth.login.config");
        if (System.getProperty("java.security.auth.login.config") == null ||
                loginConfig != null) {
            try {
                URL url = filterConfig.getServletContext().getResource(
                        (loginConfig == null) ? "/WEB-INF/login.conf" :
                                loginConfig);
                if (url != null) loginConfig = url.toExternalForm();
            } catch (MalformedURLException ex) { }
            if (loginConfig != null) {
                System.setProperty("java.security.auth.login.config",
                        loginConfig);
            } else {
                URL url = getClass().getResource("/jcifs/http/login.conf");
                if (url != null) {
                    System.setProperty("java.security.auth.login.config",
                            url.toExternalForm());
                }
            }
        }
        String useSubjectCredsOnly = filterConfig.getInitParameter(
                "javax.security.auth.useSubjectCredsOnly");
        if (useSubjectCredsOnly != null) {
            System.setProperty("javax.security.auth.useSubjectCredsOnly",
                    useSubjectCredsOnly);
        }
        String krbConf =
                filterConfig.getInitParameter("java.security.krb5.conf");
        if (krbConf != null) {
            try {
                String realPath =
                        filterConfig.getServletContext().getRealPath(krbConf);
                if (realPath != null && (new File(realPath)).isFile()) {
                    krbConf = realPath;
                }
            } catch (Exception ex) { }
            System.setProperty("java.security.krb5.conf", krbConf);
        }
        String krbRealm =
                filterConfig.getInitParameter("java.security.krb5.realm");
        if (krbRealm != null) {
            System.setProperty("java.security.krb5.realm", krbRealm);
        }
        String krbKdc = filterConfig.getInitParameter("java.security.krb5.kdc");
        if (krbKdc != null) {
            System.setProperty("java.security.krb5.kdc", krbKdc);
        }
        defaultDomain = Config.getProperty("jcifs.smb.client.domain");
        domainController = Config.getProperty("jcifs.http.domainController");
        if (domainController == null) {
            domainController = defaultDomain;
            loadBalance = Config.getBoolean("jcifs.http.loadBalance", true);
        }
        enableNegotiate =
                Config.getBoolean("jcifs.http.enableNegotiate", false);
        enableBasic = Config.getBoolean("jcifs.http.enableBasic", false);
        insecureBasic = Config.getBoolean("jcifs.http.insecureBasic", false);
        realm = Config.getProperty("jcifs.http.basicRealm");
        if (realm == null) realm = "jCIFS";
    }

    public void destroy() {
        filterConfig = null;
    }

    public void setFilterConfig(FilterConfig filterConfig) {
        try {
            init(filterConfig);
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    public FilterConfig getFilterConfig() {
        return filterConfig;
    }

    public void doFilter(ServletRequest request, ServletResponse response,
            FilterChain chain) throws IOException, ServletException {
        HttpServletRequest req = (HttpServletRequest) request;
        HttpServletResponse resp = (HttpServletResponse) response;
        UniAddress dc;
        Principal principal = null;
        String authType = null;
        String msg = req.getHeader("Authorization");
        boolean offerBasic = enableBasic && (insecureBasic || req.isSecure());
        if (msg != null && (msg.regionMatches(true, 0, "Negotiate ", 0, 10) ||
                msg.regionMatches(true, 0, "NTLM ", 0, 5) || (offerBasic &&
                        msg.regionMatches(true, 0, "Basic ", 0, 6)))) {
            if (loadBalance) {
                dc = new UniAddress(NbtAddress.getByName(domainController,
                        0x1C, null));
            } else {
                dc = UniAddress.getByName(domainController, true);
            }
            authType = msg.regionMatches(true, 0, "Negotiate ", 0, 10) ?
                    HTTP_NEGOTIATE : msg.regionMatches(true, 0, "NTLM ", 0, 5) ?
                            HTTP_NTLM : HTTP_BASIC;
            if (HTTP_NEGOTIATE.equals(authType)) {
                principal = Negotiate.authenticate(req, resp);
                if (principal == null) return;
                if (!(principal instanceof NtlmPasswordAuthentication)) {
                    req.getSession().setAttribute("jcifs.http.principal",
                            principal);
                    chain.doFilter(new AuthenticatedRequest(req, principal,
                            authType), response);
                    return;
                }
            } else if (HTTP_NTLM.equals(authType) ||
                    HTTP_BASIC.equals(authType)) {
                if (HTTP_NTLM.equals(authType)) {
                    byte[] challenge = SmbSession.getChallenge(dc);
                    if ((principal = NtlmSsp.authenticate(req, resp,
                            challenge)) == null) {
                        return;
                    }
                } else {
                    String auth = new String(Base64.decode(msg.substring(6)),
                            "US-ASCII");
                    int index = auth.indexOf(':');
                    String user = (index != -1) ?
                            auth.substring(0, index) : auth;
                    String password = (index != -1) ?
                            auth.substring(index + 1) : "";
                    index = user.indexOf('\\');
                    if (index == -1) index = user.indexOf('/');
                    String domain = (index != -1) ? user.substring(0, index) :
                            defaultDomain;
                    user = (index != -1) ? user.substring(index + 1) : user;
                    principal = new NtlmPasswordAuthentication(domain, user,
                            password);
                }
            }
            try {
                SmbSession.logon(dc,
                        (NtlmPasswordAuthentication) principal);
            } catch (SmbAuthException sae) {
                if (sae.getNtStatus() == sae.NT_STATUS_ACCESS_VIOLATION) {
                    HttpSession ssn = req.getSession(false);
                    if (ssn != null) {
                        ssn.removeAttribute("jcifs.http.principal");
                    }
                }
                if (enableNegotiate) {
                    resp.setHeader("WWW-Authenticate", "Negotiate");
                    resp.addHeader("WWW-Authenticate", "NTLM");
                } else {
                    resp.setHeader("WWW-Authenticate", "NTLM");
                }
                if (offerBasic) {
                    resp.addHeader("WWW-Authenticate", "Basic realm=\"" +
                            realm + "\"");
                }
                resp.setHeader("Connection", "close");
                resp.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
                resp.flushBuffer();
                return;
            }
            HttpSession ssn = req.getSession();
            ssn.setAttribute("jcifs.http.principal", principal);
        } else {
            HttpSession ssn = req.getSession(false);
            if (ssn == null || (principal = (Principal)
                    ssn.getAttribute("jcifs.http.principal")) == null) {
                if (enableNegotiate) {
                    resp.setHeader("WWW-Authenticate", "Negotiate");
                    resp.addHeader("WWW-Authenticate", "NTLM");
                } else {
                    resp.setHeader("WWW-Authenticate", "NTLM");
                }
                if (offerBasic) {
                    resp.addHeader("WWW-Authenticate", "Basic realm=\"" +
                            realm + "\"");
                }
                resp.setHeader("Connection", "close");
                resp.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
                resp.flushBuffer();
                return;
            }
        }
        chain.doFilter(new AuthenticatedRequest(req, principal, authType),
                response );
    }

}

