/* jcifs smb client library in Java
 * Copyright (C) 2004  "Michael B. Allen" <jcifs at samba dot org>
 *                   "Eric Glass" <jcifs at samba dot org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.spnego;

import java.net.UnknownHostException;

import java.security.Principal;
import java.security.PrivilegedExceptionAction;

import java.util.Properties;

import javax.security.auth.Subject;

import javax.security.auth.kerberos.KerberosPrincipal;

import javax.security.auth.login.LoginContext;

import jcifs.Config;
import jcifs.UniAddress;

import jcifs.netbios.NbtAddress;

import jcifs.ntlmssp.NtlmFlags;
import jcifs.ntlmssp.Type1Message;
import jcifs.ntlmssp.Type2Message;
import jcifs.ntlmssp.Type3Message;

import jcifs.smb.NtlmPasswordAuthentication;
import jcifs.smb.SmbSession;

import org.ietf.jgss.GSSContext;
import org.ietf.jgss.GSSCredential;
import org.ietf.jgss.GSSManager;
import org.ietf.jgss.GSSName;
import org.ietf.jgss.Oid;

public class Authentication {

    private static final byte[] NTLMSSP_SIGNATURE = new byte[] {
        (byte) 'N', (byte) 'T', (byte) 'L', (byte) 'M',
        (byte) 'S', (byte) 'S', (byte) 'P', (byte) 0
    };

    private Properties properties = new Properties();

    private boolean client;

    private byte[] challenge;

    private byte[] nextToken;

    private Subject subject;

    private Principal principal;

    private GSSContext context;

    public Authentication() {
        this(null);
    }

    public Authentication(Properties properties) {
        setProperties(properties);
    }

    public byte[] getNextToken() {
        return nextToken;
    }

    public Principal getPrincipal() {
        return principal;
    }

    public Properties getProperties() {
        return properties;
    }

    public void setProperties(Properties properties) {
        this.properties = (properties != null) ? properties : new Properties();
    }

    public String getProperty(String property) {
        String value = getProperties().getProperty(property);
        if (value == null) value = Config.getProperty(property);
        return (value != null) ? value : System.getProperty(property);
    }

    public void setProperty(String property, String value) {
        if (value != null) {
            getProperties().setProperty(property, value);
        } else {
            getProperties().remove(property);
        }
    }

    public void init(String[] mechanisms) throws AuthenticationException {
        reset();
        client = true;
        if (mechanisms == null || mechanisms.length == 0) {
            String unicode = getProperty("jcifs.smb.client.useUnicode");
            boolean useUnicode = (unicode != null) ?
                    Boolean.valueOf(unicode).booleanValue() : true;
            int flags = NtlmFlags.NTLMSSP_NEGOTIATE_NTLM | (useUnicode ?
                    NtlmFlags.NTLMSSP_NEGOTIATE_UNICODE :
                            NtlmFlags.NTLMSSP_NEGOTIATE_OEM);
            String domain = getProperty("jcifs.smb.client.domain");
            String workstation = null;
            try {
                workstation = NbtAddress.getLocalHost().getHostName();
            } catch (UnknownHostException ex) { }
            nextToken = new Type1Message(flags, domain,
                    workstation).toByteArray();
            return;
        } else {
            int contextFlags = 0;
            byte[] mechanismToken = null;
            String initialMechanism = mechanisms[0];
            if (SpnegoConstants.NTLMSSP_MECHANISM.equals(initialMechanism)) {
                String unicode = getProperty("jcifs.smb.client.useUnicode");
                boolean useUnicode = (unicode != null) ?
                        Boolean.valueOf(unicode).booleanValue() : true;
                int flags = NtlmFlags.NTLMSSP_NEGOTIATE_NTLM | (useUnicode ?
                        NtlmFlags.NTLMSSP_NEGOTIATE_UNICODE :
                                NtlmFlags.NTLMSSP_NEGOTIATE_OEM);
                String domain = getProperty("jcifs.smb.client.domain");
                String workstation = null;
                try {
                    workstation = NbtAddress.getLocalHost().getHostName();
                } catch (UnknownHostException ex) { }
                mechanismToken = new Type1Message(flags, domain,
                        workstation).toByteArray();
            } else if (SpnegoConstants.KERBEROS_MECHANISM.equals(
                    initialMechanism) ||
                            SpnegoConstants.LEGACY_KERBEROS_MECHANISM.equals(
                                    initialMechanism)) {
                ClientAction action = new ClientAction(null);
                String useSubjectCredsOnly = getProperty(
                        "javax.security.auth.useSubjectCredsOnly");
                boolean subjectOnly = (useSubjectCredsOnly == null) ? true :
                        Boolean.valueOf(useSubjectCredsOnly).booleanValue();
                if (subjectOnly) {
                    try {
                        String name = getProperty(
                                "javax.security.auth.login.name");
                        String password = getProperty(
                                "javax.security.auth.login.password");
                        LoginContext loginContext = new LoginContext(
                                "jcifs.spnego.initiate",
                                        new SpnegoLoginHandler(name, password));
                        loginContext.login();
                        subject = loginContext.getSubject();
                    } catch (Exception ex) {
                        throw new AuthenticationException("Unable to login: " +
                                ex, ex);
                    }
                }
                try {
                    mechanismToken = (byte[]) ((subject != null) ?
                            Subject.doAsPrivileged(subject, action, null) :
                                    action.run());
                } catch (Exception ex) {
                    throw new AuthenticationException(
                            "Error processing token: " + ex, ex);
                }
                if (context.getCredDelegState()) {
                    contextFlags |= NegTokenInit.DELEGATION;
                }
                if (context.getMutualAuthState()) {
                    contextFlags |= NegTokenInit.MUTUAL_AUTHENTICATION;
                }
                if (context.getReplayDetState()) {
                    contextFlags |= NegTokenInit.REPLAY_DETECTION;
                }
                if (context.getSequenceDetState()) {
                    contextFlags |= NegTokenInit.SEQUENCE_CHECKING;
                }
                if (context.getAnonymityState()) {
                    contextFlags |= NegTokenInit.ANONYMITY;
                }
                if (context.getConfState()) {
                    contextFlags |= NegTokenInit.CONFIDENTIALITY;
                }
                if (context.getIntegState()) {
                    contextFlags |= NegTokenInit.INTEGRITY;
                }
            } else {
                throw new UnsupportedMechanismException(
                        "Unsupported initial mechanism: " + initialMechanism);
            }
            nextToken = new NegTokenInit(mechanisms, contextFlags,
                    mechanismToken, null).toByteArray();
        }
    }

    public void reset() {
        client = false;
        challenge = null;
        nextToken = null;
        subject = null;
        principal = null;
        context = null;
    }

    public void process(byte[] token) throws AuthenticationException {
        if (token == null) throw new NullPointerException("Null token.");
        if (getPrincipal() != null) {
            throw new IllegalStateException("Context already completed.");
        }
        if (isNtlm(token)) {
            processNtlm(token);
        } else {
            processSpnego(token);
        }
    }

    private void processNtlm(byte[] token) throws AuthenticationException {
        try {
            switch (token[8]) {
            case 1:
                if (client) {
                    throw new AuthenticationException(
                            "NTLM Type 1 message received by client.");
                }
                boolean loadBalance = false;
                String domainController =
                        getProperty("jcifs.http.domainController");
                if (domainController == null) {
                    domainController = getProperty("jcifs.smb.client.domain");
                    String balance = getProperty("jcifs.http.loadBalance");
                    loadBalance = (balance != null) ?
                            Boolean.valueOf(balance).booleanValue() : true;

                }
                UniAddress dc = loadBalance ? new UniAddress(
                        NbtAddress.getByName(domainController, 0x1c, null)) :
                                UniAddress.getByName(domainController, true);
                challenge = SmbSession.getChallenge(dc);
                nextToken = new Type2Message(new Type1Message(token),
                        challenge, null).toByteArray();
                break;
            case 2:
                if (!client) {
                    throw new AuthenticationException(
                            "NTLM Type 2 message received by server.");
                }
                String workstation = null;
                try {
                    workstation = NbtAddress.getLocalHost().getHostName();
                } catch (UnknownHostException ex) { }
                nextToken = new Type3Message(new Type2Message(token),
                        getProperty("jcifs.smb.client.password"),
                        getProperty("jcifs.smb.client.domain"),
                        getProperty("jcifs.smb.client.username"),
                        workstation).toByteArray();
                break;
            case 3:
                if (client) {
                    throw new AuthenticationException(
                            "NTLM Type 3 message received by client.");
                }
                Type3Message type3 = new Type3Message(token);
                byte[] lmResponse = type3.getLMResponse();
                if (lmResponse == null) lmResponse = new byte[0];
                byte[] ntResponse = type3.getNTResponse();
                if (ntResponse == null) ntResponse = new byte[0];
                principal = new NtlmPasswordAuthentication(type3.getDomain(),
                        type3.getUser(), challenge, lmResponse, ntResponse);
                nextToken = null;
                break;
            default:
                throw new AuthenticationException(
                        "Unrecognized NTLM Token Type: " + token[8]);
            }
        } catch (AuthenticationException ex) {
            throw ex;
        } catch (Exception ex) {
            throw new AuthenticationException(
                    "Error performing Kerberos authentication: " + ex, ex);
        }
    }

    private void processSpnego(byte[] token) throws AuthenticationException {
        try {
            String mechanism;
            byte[] mechanismToken;
            switch (token[0]) {
            case (byte) 0x60:
                NegTokenInit initToken = new NegTokenInit(token);
                mechanismToken = initToken.getMechanismToken();
                mechanism = (mechanismToken != null) ?
                        initToken.getMechanisms()[0] : null;
                break;
            case (byte) 0xa1:
                NegTokenTarg targToken = new NegTokenTarg(token);
                mechanismToken = targToken.getMechanismToken();
                mechanism = targToken.getMechanism();
                break;
            default:
                throw new AuthenticationException("Unrecognized SPNEGO Token.");
            }
            if (isNtlm(mechanismToken)) {
                processNtlm(mechanismToken);
                if (getNextToken() == null) return;
                int result = (getPrincipal() == null) ?
                        NegTokenTarg.ACCEPT_INCOMPLETE :
                                NegTokenTarg.ACCEPT_COMPLETED;
                nextToken = new NegTokenTarg(result,
                        SpnegoConstants.NTLMSSP_MECHANISM, mechanismToken,
                                null).toByteArray();
            } else {
                try {
                    processKerberos(mechanism, mechanismToken);
                } catch (UnsupportedMechanismException ex) {
                    nextToken = new NegTokenTarg(NegTokenTarg.REJECTED, null,
                            null, null).toByteArray();
                }
            }
        } catch (AuthenticationException ex) {
            throw ex;
        } catch (Exception ex) {
            throw new AuthenticationException(
                    "Error performing SPNEGO negotiation: " + ex, ex);
        }
    }

    private void processKerberos(String mechanism, byte[] token)
            throws AuthenticationException {
        try {
            if (client) {
                ClientAction action = new ClientAction(token);
                token = (byte[]) ((subject != null) ?
                        Subject.doAsPrivileged(subject, action, null) :
                                action.run());
                int result;
                if (context.isEstablished()) {
                    result = NegTokenTarg.ACCEPT_COMPLETED;
                    principal = new KerberosPrincipal(
                            context.getSrcName().toString());
                } else {
                    result = NegTokenTarg.ACCEPT_INCOMPLETE;
                }
                nextToken = new NegTokenTarg(result, mechanism, token,
                        null).toByteArray();
            } else {
                ServerAction action = new ServerAction(token);
                if (subject == null) {
                    String useSubjectCredsOnly = getProperty(
                            "javax.security.auth.useSubjectCredsOnly");
                    boolean subjectOnly = (useSubjectCredsOnly == null) ? true :
                            Boolean.valueOf(useSubjectCredsOnly).booleanValue();
                    if (subjectOnly) {
                        String name = getProperty(
                                "jcifs.spnego.servicePrincipal");
                        String password =
                                getProperty("jcifs.spnego.servicePassword");
                        LoginContext loginContext = new LoginContext(
                                "jcifs.spnego.accept",
                                        new SpnegoLoginHandler(name, password));
                        loginContext.login();
                        subject = loginContext.getSubject();
                    }
                }
                token = (byte[]) ((subject != null) ?
                        Subject.doAsPrivileged(subject, action, null) :
                                action.run());
                int result;
                if (context.isEstablished()) {
                    result = NegTokenTarg.ACCEPT_COMPLETED;
                    principal = new KerberosPrincipal(
                            context.getSrcName().toString());
                } else {
                    result = NegTokenTarg.ACCEPT_INCOMPLETE;
                }
                nextToken = new NegTokenTarg(result, mechanism, token,
                        null).toByteArray();
            }
        } catch (Exception ex) {
            throw new AuthenticationException(
                    "Error performing Kerberos authentication: " + ex, ex);
        }
    }

    private static boolean isNtlm(byte[] token) {
        if (token == null || token.length < 8) return false;
        for (int i = 0; i < 8; i++) {
            if (NTLMSSP_SIGNATURE[i] != token[i]) return false;
        }
        return true;
    }

    private class ClientAction implements PrivilegedExceptionAction {

        private byte[] token;

        public ClientAction(byte[] token) {
            this.token = (token != null) ? token : new byte[0];
        }

        public Object run() throws Exception {
            if (Authentication.this.context == null) {
                Oid kerberos = new Oid(SpnegoConstants.KERBEROS_MECHANISM);
                GSSManager gss = GSSManager.getInstance();
                GSSName servicePrincipal = gss.createName(
                        getProperty("jcifs.spnego.servicePrincipal"), null);
                GSSCredential credential = null;
                String user = getProperty("javax.security.auth.login.name");
                if (user != null) {
                    credential = gss.createCredential(
                            gss.createName(user, null),
                                    GSSContext.DEFAULT_LIFETIME, kerberos,
                                            GSSCredential.INITIATE_ONLY);
                }
                Authentication.this.context =
                        gss.createContext(servicePrincipal, kerberos,
                                credential, GSSContext.DEFAULT_LIFETIME);
            }
            return Authentication.this.context.initSecContext(
                    token, 0, token.length);
        }

    }

    private class ServerAction implements PrivilegedExceptionAction {

        private byte[] token;

        public ServerAction(byte[] token) {
            this.token = (token != null) ? token : new byte[0];
        }

        public Object run() throws Exception {
            if (Authentication.this.context == null) {
                Oid kerberos = new Oid(SpnegoConstants.KERBEROS_MECHANISM);
                GSSManager gss = GSSManager.getInstance();
                GSSName servicePrincipal = gss.createName(
                        getProperty("jcifs.spnego.servicePrincipal"), null);
                GSSCredential credential =
                        gss.createCredential(servicePrincipal,
                                GSSContext.DEFAULT_LIFETIME, kerberos,
                                        GSSCredential.ACCEPT_ONLY);
                Authentication.this.context =
                        gss.createContext(credential);
            }
            return Authentication.this.context.acceptSecContext(
                    token, 0, token.length);
        }

    }

}
